from typing import List

from .operation import Operation


def merge(left: List[int], right: List[int], ops: List[Operation]):
    """
    q の値に基づいてソート済みの配列を， q の値に基づいて降順にマージする
    """
    merged = list()
    len_left, len_right = len(left), len(right)
    i, j = 0, 0

    while i < len_left and j < len_right:
        if ops[left[i]].q > ops[right[j]].q:
            merged.append(left[i])
            i += 1
        else:
            merged.append(right[j])
            j += 1

    if i < len_left:
        merged.extend(left[i:])
    if j < len_right:
        merged.extend(right[j:])
    return merged



from typing import Dict, List, NewType, Tuple

from .utils.operation import Operation, create_operations
from .utils import algo

TimeInterval = NewType('TimeInterval', Tuple[int, int])


def jackson_preemptive_algorithm(r_list: List[int], p_list: List[int], q_list: List[int]) -> Tuple[List[Operation], Dict[TimeInterval, Operation], int]:
    def complete(op: Operation, alpha: List[int], schedule: Dict[TimeInterval, Operation], t: int, makespan: int) -> int:
        Cjk = t + op.p_plus
        op.C, op.p_plus = Cjk, 0
        alpha[k] += 1
        alpha[k] = alpha[k] if alpha[k] < 2 ** k else None
        schedule[(t, Cjk)] = op
        return Cjk, max(Cjk + op.q, makespan)

    def preempt(op: Operation, alpha: List[int], schedule: Dict[TimeInterval, Operation], t: int) -> int:
        next_t = sorted_ops[m].r
        op.p_plus -= next_t - t
        schedule[(t, next_t)] = op
        return next_t

    def get_max_q_from_S(k: int) -> int:
        Sk = S[k]
        max_q_idx = alpha[k]
        max_q_op = Sk[max_q_idx]
        max_q = sorted_ops[max_q_op].q
        return max_q

    unsorted_ops = create_operations(r_list, p_list, q_list)
    sorted_ops = sorted(unsorted_ops, key=lambda op: op.r)
    m, n, t, p = algo.setup_params(sorted_ops)
    alpha, S, schedule = algo.setup_container(p)

    makespan = -1
    while True:
        # Does job m+1 belong to S?
        # belong
        if m < n and sorted_ops[m].r <= t:
            # Insertion of a job in S
            for k in range(n):
                if len(S[k]) == 0:
                    algo.update_S(S, sorted_ops, alpha, k, m)
                    m += 1
                    break

        ## not belong
        else:
            # Selection of job j

            # all jobs in I are scheduled
            if all(op.is_done() for op in unsorted_ops):
                for op in unsorted_ops:
                    op.p_plus = op.p
                return unsorted_ops, schedule, makespan

            # all jobs in S are scheduled
            is_scheduled_for_all_jobs = list()
            for Sk in S:
                for idx in Sk:
                    is_scheduled_for_all_jobs.append(sorted_ops[idx].is_done())
            if all(is_scheduled_for_all_jobs):
                t = sorted_ops[m].r
                continue

            k = max([k for k in range(p + 1) if alpha[k] is not None], key=get_max_q_from_S)
            jk = S[k][alpha[k]]

            op = sorted_ops[jk]
            if m >= n:
                t, makespan = complete(op, alpha, schedule, t, makespan)
            else:
                if t + op.p_plus <= sorted_ops[m].r:
                    t, makespan = complete(op, alpha, schedule, t, makespan)
                else:
                    t = preempt(op, alpha, schedule, t)



import math
from typing import List

from .merge import merge
from .operation import Operation


def update_S(S: List[List[int]], ops: List[Operation], alpha: List[int], k0: int, m: int):
    """
    S の更新をする.
    参照渡しにより更新するので注意
    """
    scheduled, asc_not_scheduled = list(), [m]
    for i in range(k0):
        # NOTE: αはリストが空の場合 or 全てがスケジュール済みの場合の両方で None になるので,
        # スケジュール済みとそうでないものを分けるために, lenで取得する必要がある
        partition_job_idx = len(S[i]) if alpha[i] is None else alpha[i]
        scheduled += S[i][:partition_job_idx]
        asc_not_scheduled = merge(asc_not_scheduled, S[i][partition_job_idx:], ops)
        S[i] = list()
        alpha[i] = None

    alpha[k0] = len(scheduled)
    S[k0] = scheduled + asc_not_scheduled


def setup_params(sorted_ops: List[Operation]):
    m, n = 0, len(sorted_ops)
    t = sorted_ops[0].r
    p = math.floor(math.log2(n))
    return m, n, t, p


def setup_container(p: int):
    alpha = [None for _ in range(p + 1)]
    S = [list() for _ in range(p + 1)]
    schedule = dict()
    return alpha, S, schedule



# -*- coding: utf-8 -*-
from __future__ import annotations
from jsp_exp import Operation as BaseOperation
from typing import Union


# 基本のOperationを拡張する
class Operation(BaseOperation):
    def __init__(self, processed_by: Union[int, str] = 'Machine', name: str = 'Operation', r: Optional[int] = None, p: Optional[int] = None, q: Optional[int] = None):
        super().__init__(processed_by, name, r, p, q)
        self.__p_plus = self.p

    @property
    def p_plus(self) -> int:
        return self.__p_plus

    @p_plus.setter
    def p_plus(self, new_val: int) -> None:
        """Validate new p_plus value is valid or not.

        Args:
            new_val (int): new p_plus value. If it is valid, it must be a positive integer.
        """
        assert new_val >= 0, f"Operation '{self.name}' is overscheduled. Please check your code."
        self.__p_plus = new_val

    def is_done(self) -> bool:
        """Return the job is already completed or not.

        Returns:
            bool: If job is already completed, then return True. Otherwise False.
        """
        return self.p_plus == 0


def create_operations(r_list: List[int], p_list: List[int], q_list: List[int]):
    unsorted_ops = [
        Operation(processed_by='M', name=f"O{i+1}", r=r, p=p, q=q)
        for i, (r, p, q) in enumerate(zip(r_list, p_list, q_list))
    ]
    return unsorted_ops



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jps_algo', 'jps_algo.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'jps-algo',
    'version': '0.1.2',
    'description': "Jackson's Preemptive Scheduleを解くためのアルゴリズムが実装されている",
    'long_description': None,
    'author': '鈴木貴大',
    'author_email': 'merioda.seven.24@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tkp0331/jps-algo',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.9,<4.0.0',
}


setup(**setup_kwargs)
